package jp.co.sds_corp.tool.test;

import	java.text.SimpleDateFormat;
import	java.util.ArrayList;
import	java.util.Calendar;
import	java.util.Date;
import	java.util.List;
import	java.util.Timer;
import	java.util.TimerTask;
import	java.util.concurrent.BlockingQueue;
import	java.util.concurrent.LinkedBlockingQueue;
import	java.util.concurrent.TimeUnit;

import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

public class StressTester<T>
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	StressExecutor<T> executor;
	boolean initPool;
	BlockingQueue<T> queue;
	int	poolSize = 10;
	List<ExecThread> listThread;
	int caseCount;
	int loopCount;
	int casePos;
	int loopPos;
	Timer timer;
	KickTask task;
	List<T> listCase;
	Object lock = new Object();
	Date	sendTime;
	SimpleDateFormat sdf = new SimpleDateFormat( "yyyy/MM/dd HH:mm.ss" );
	int count = 0;
	StressResult total = new StressResult();
	ShutdownThread shut;
	boolean waitResult = true;
	boolean finished = false;

	public StressTester( StressExecutor<T> executor ) {
		this.executor = executor;
		shut = new ShutdownThread();
		Runtime.getRuntime().addShutdownHook( shut );
	}

	public void test( List<T> list, long interval, int count )
		throws Exception
	{
		if ( interval == 0L ) {
			logger.error( this, "interval must >0" );
			return;
		}

		if ( !isPoolInitialized() ) {
			initThreadPool();
		}

		boolean canTest = false;
		for (int retry = 0; retry < 100; retry++) {
			boolean running = false;
			if ( timer != null || !queue.isEmpty()) {
				running = true;
			} else {
				for ( ExecThread thr : listThread ) {
					if ( thr.getResult().isRunning()) {
						running = true;
						break;
					}
				}
			}
			if (!running) {
				canTest = true;
				break;
			}
			// logger.trace("==== retry");
			Thread.sleep(50);
		}
		if (!canTest) {
			logger.warn( this, "===== running." );
			return;
		}

		synchronized( this ) {
			listCase = list;
			caseCount = listCase.size();
			casePos = -1;
			loopCount = count;
			loopPos = 0;
			finished = false;
			for ( ExecThread thr : listThread ) {
				thr.getResult().clear();
			}
			total.clear();
		}

		for ( ExecThread thr : listThread ) {
			thr.waitStart();
			logger.trace( this, "thread " + thr.no + " started." );
		}

		logger.debug( this, "start test interval=" + interval + " count=" + count );
		timer = new Timer();
		task = new KickTask();
		timer.scheduleAtFixedRate( task, 0, interval );

		if ( waitResult ) {
			synchronized ( lock ) {
				for ( ;; ) {
					if (finished) {
						break;
					}
					try {
						lock.wait();
					} catch (InterruptedException iex) {
						break;
					}
				}
			}
			timer.cancel();
			timer = null;
		}
	}

	public void breakTest() {
		if (timer != null) {
			timer.cancel();
			timer = null;
		}
		queue.clear();
		if ( waitResult ) {
			synchronized ( lock ) {
				finished = true;
				lock.notifyAll();
			}
		}
	}

	public void setWaitResult( boolean var ) {
		this.waitResult = var;
	}

	void initThreadPool() {
		logger.info( this, "init ThreadPool size=" + poolSize );
		queue = new LinkedBlockingQueue<T>();
		listThread = new ArrayList<ExecThread>();
		for ( int i = 0; i < poolSize; i++ ) {
			ExecThread thread = new ExecThread( i );
			listThread.add( thread );
			thread.start();
		}
		setPoolInitialized(true);
	}

	class ShutdownThread extends Thread {
		public void run() {
			try {
				StressTester.this.stop();
			} catch (Exception ex) {
				logger.error(this, "stop error", ex);
			}
		}
	}

	class KickTask extends TimerTask {
		public void run() {
			int pos = getCasePos();
			if ( pos < 0 ) {
				cancel();
				synchronized ( lock ) {
					finished = true;
					lock.notifyAll();
				}
				if ( !waitResult && timer != null ) {
					timer.cancel();
					timer = null;
				}
				return;
			}
			T testData = listCase().get( pos );
			try {
				queue.put( testData );
			} catch ( InterruptedException ex ) {
			}
		}
	}

	synchronized List<T> listCase() {
		return this.listCase;
	}

	synchronized int getCasePos() {
		casePos++;
		if ( casePos >= caseCount ) {
			casePos = 0;
			loopPos++;
		}
		if ( loopCount > 0 && loopPos >= loopCount ) {
			casePos = -1;
		}
		return casePos;
	}

	class ExecThread extends Thread {
		int no;
		boolean start = false;
		boolean stop = false;
		StressResult result;

		ExecThread( int no ) {
			this.no = no;
			setName("StressTester-" + no);
			setResult(new StressResult());
		}
		public void run() {
			logger.debug( this, "start thread " + no );
			try {
				executor.startThread();
			} catch ( Exception ex ) {
				logger.error( this, "start error", ex );
				return;
			}
			setStart( true );
			for ( ;; ) {
				if ( isStop()) {
					break;
				}
				T testData;
				try {
					testData = queue.poll( 1000, TimeUnit.MILLISECONDS );
				} catch ( InterruptedException iex ) {
					logger.debug( this, "" + no + " interrupted" );
					break;
				}
				if ( testData == null ) {
					continue;
				}
				StressResult result = getResult();
				try {
					logger.trace( this, "exec " + no );
					result.start();
					executor.test( testData );
					result.end();
				} catch ( Exception ex ) {
					result.end();
					logger.error( this, "test error", ex );
				}
				addCount();
			}
			try {
				executor.stopThread();
			} catch ( Exception ex ) {
				logger.error( this, "stop error", ex );
			}
			logger.debug( this, "thread " + no + " stopped" );
		}

		public synchronized void waitStart() {
			for ( ;; ) {
				if ( start ) {
					break;
				}
				try {
					wait();
				} catch ( InterruptedException ex ) {
					break;
				}
			}
		}
		public synchronized void setStart( boolean var ) {
			start = true;
			notifyAll();
		}

		public synchronized void setStop( boolean var ) {
			this.stop = var;
		}
		public synchronized boolean isStop() {
			return this.stop;
		}
		public synchronized void setResult(StressResult var) {
			result = var;
		}
		public synchronized StressResult getResult() {
			return result;
		}
	}

	synchronized int getCount() {
		return this.count;
	}

	void addCount() {
		Calendar nowCal = Calendar.getInstance();
		nowCal.clear( Calendar.MILLISECOND );
		Date now = nowCal.getTime();

		boolean print = false;
		int pCount = getCount();
		Date pDate;
		synchronized ( this ) {
			pDate = sendTime;
			count++;
			if ( sendTime == null ) {
				sendTime = now;
			} else {
				if ( !sendTime.equals( now )) {
					print = true;
					sendTime = now;
					count = 0;
				}
			}
		}
		if ( print ) {
			logger.detail( this, sdf.format( pDate ) + " "
									+ pCount + " messages. "
									+ queue.size() + " left." );
		}
	}

	synchronized boolean isPoolInitialized() {
		return this.initPool;
	}
	synchronized void setPoolInitialized(boolean var) {
		this.initPool = var;
	}

	public void setPoolSize( int var ) {
		this.poolSize = var;
		setPoolInitialized(false);
	}
	public int getPoolSize() {
		return this.poolSize;
	}

	public StressResult getResult( int threadno ) {
		ExecThread thr = listThread.get( threadno );
		StressResult result = thr.getResult();
		return result;
	}

	public StressResult getTotalResult() {
		return total;
	}

	public List<String> getResponse() {
		List<String> ret = new ArrayList<String>();
		if ( !isPoolInitialized() ) {
			return ret;
		}
		Calendar now = Calendar.getInstance();
		ret.add( sdf.format( now.getTime()) + " "
							+ getCount() + " messages. "
							+ queue.size() + " left." );
		ret.add( "no) count,min,max,total");
		for ( int i = 0; i < poolSize; i++ ) {
			StressResult result = getResult( i );
			total.setCount( total.getCount() + result.getCount());
			if ( total.getMinTime() > result.getMinTime()) {
				total.setMinTime( result.getMinTime());
			}
			if ( total.getMaxTime() < result.getMaxTime()) {
				total.setMaxTime( result.getMaxTime());
			}
			total.setTotalTime( total.getTotalTime() + result.getTotalTime());

			ret.add( "response " + i + ") "
				+ result.getCount() + ","
				+ (result.getMinTime() == Long.MAX_VALUE ?
					"-" : result.getMinTime()) + ","
				+ (result.getMaxTime() == Long.MIN_VALUE ?
					"-" : result.getMaxTime()) + ","
				+ result.getTotalTime() );
		}
		ret.add( "total result: "
				+ total.getCount() + ","
				+ total.getMinTime() + ","
				+ total.getMaxTime() + ","
				+ total.getTotalTime() );
		return ret;
	}

	void showResponse() {
		List<String> list = getResponse();
		for (String msg : list) {
			logger.detail( this, msg );
		}
	}

	public void stop() throws Exception {
		if ( timer != null ) {
			timer.cancel();
			timer = null;
		}
		if ( isPoolInitialized() ) {
			for ( ExecThread thr : listThread ) {
				thr.setStop( true );
			}
			for ( ExecThread thr : listThread ) {
				thr.join();
			}
			showResponse();
			setPoolInitialized(false);
		}
	}

}
